#include <SPI.h>
#include <math.h>

const int slaveSelectPin = 10;
const int writeEnable = 0b00011000;
const int firstByte = 0b00000111;

const int readPin = A5;
const int vinPin = A0;

int b1 = 5;
int b2 = 200;


void setup() {

  Serial.begin(9600);

  pinMode(slaveSelectPin, OUTPUT);
  SPI.begin();
  SPI.setBitOrder(MSBFIRST);
  SPI.setDataMode(SPI_MODE1);

  // Enable writing to RDAC
  digitalWrite(slaveSelectPin, LOW); 
 
  SPI.transfer(writeEnable); // Write first byte
  SPI.transfer(firstByte); // Write second byte
  
  digitalWrite(slaveSelectPin, HIGH); delay(100);

  // Initialize digipot to output just under 10V in the converter.
  digitalPotWrite(b1, b2);

}

void loop() {

  digitalPotWrite(b1, b2);
  
  // Current Reading
  int val = analogRead(readPin);
  double current = val * (5.0 / 1023);

  // Calculate the theoretical voltage at this current
  // Use (5 - current) instead of just (current) to mimic open circuit and short circuit behavior
  double vCalc = (1 / 0.02568) * log(((5 - current) / 1.0937) + 1) / 3.5; // /36
  if (vCalc <= 1.23) {
    vCalc = 1.23;
  }

  // Voltage Reading
  int vin = analogRead(vinPin);
  double voltage = vin * (5.0 / 1023 / 0.192);

  // If we're reading higher voltage than we expect by the diode equation:
  if (voltage > vCalc) {
    
    // Reduce the digipot setting by one increment ~ 78.125 Ohms
    // This reduction corresponds to ~96 mV
    if ((b1 >= 4) && (b2 >= 4)) {
      b2 = b2 - 3;
    }
    

    // If we get to zero on byte 2, decrement byte 1 and reset byte 2 to 255
    if (b2 <= 4) {
      if (b1 > 4) {
        b1 = --b1;
        b2 = 255;
      }
      // If minimum value
      else if (b1 == 4) {
        b1 = b1;
        b2 = b2;
      }
    }
  }

  // If we're reading lower voltage than we expect by the diode equation:
  else if (voltage < vCalc) {

    // Increment the digipot by one bit ~ 78.125 Ohms
    // This increase corresponds to ~96 mV
    // Increment only if we're not already in the maximum digipot setting
    if (((b1 <= 6) && (b2 <= 255))) {
      // To prevent open short circuit voltage from going to maximum
      if (!((vCalc == 1.23) && (voltage < vCalc))) {
        b2 = b2 + 3;
      }
    }

    // If we get to 255 on byte 2, increment byte 1 and reset byte 2 to 0
    if (b2>255) {
      if (b1 < 6) {
        b1 = ++b1;
        b2 = 4;
      }
      // If maximum value
      else if ( b1 == 6) {
        b1 = b1;
        b2 = 215;
      }
    }
  }

  Serial.print("\n vCalc = ");
  Serial.print(vCalc); // Calculated expected voltage
  Serial.print("\n Measured voltage = ");
  Serial.print(voltage); // Measured at the converter output
  Serial.print("\n Measured current = ");
  Serial.print(current);  // Desired voltage at this current
  Serial.print("\n b1 = ");
  Serial.print(b1);  // Desired voltage at this current   
  Serial.print("\n b2 = ");
  Serial.print(b2);  // Desired voltage at this current      
}

void digitalPotWrite(int byte1, int byte2) {
  digitalWrite(slaveSelectPin, LOW); 
 
  SPI.transfer(byte1); // Write first byte
  SPI.transfer(byte2); // Write second byte
  
  digitalWrite(slaveSelectPin, HIGH); delay(5);
}
